/*============================================================================
 *      MODICON, Inc., 1 High Street, North Andover, MA 01845, USA
 *      Copyright (c) $Date: 95/01/17 19:03:27 $ - All rights reserved.
 *
 *      No part of this document may be reproduced in any form
 *      without the express written consent of MODICON, Inc.
 *
 *=fdoc=======================================================================
 *
 * File:        netlib.c
 *
 * Functions:
 * Name                 	Description
 * -------------------- 	-----------
 *	ncb_reset()					Reset MB+ driver and hardware.
 *	ncb_status()				Get board status.
 *	ncb_open()					Open a path.
 *	ncb_close()					Close a path.
 *	ncb_send()					Send data on an opened path.
 *	ncb_receive()				Receive data.
 *	ncb_send_datagram()		Send global data
 *	ncb_receive_datagram()	Receive global data
 *	ncb_cancel()				Cancel a MB+ request.
 *	ncb_sa85off()				Take driver offline
 *	ncb_set_slave_login()	Set login status
 *	ncd_set_sw_interrupt()	Set SW interrupt for accessing driver 
 *	exec_netbios()				Trap to the driver.
 *                     
 *
 * Notes:		Contains source for the DOS and OS/2 static libraries as
 *					well as the Windows 3.X DLL.
 *					The netlib.c routines are used by all the example programs.
 *					Must be compiled Large model only 
 *
 * Author:	
 *
 * RCSID:
 *
 * RCS:
 * $Revision: 1.5 $
 * $State: R3_41 $
 * $Locker: $
 *
 * Revision History Log:
 *
 * 01/06/95	DP	  Introduced MAX_ADAPTERS #define. Return an error if the
 *               GlobalDosAlloc fails in the DLL version of exec_netbios.
 * 
 * 04/14/94	DP	  Merged DOS and OS/2 with Windows DLL code.
 *
 *	10/22/93 RAE -Added OS/2 ncb_close function which doesn't reference the ncb
 *	              after returning from exec_netbios.
 *	10/19/92 RAE -Added Windows 3.x awareness under DOS exec_netbios.
 *	              Fixed ncb_open for slave paths, it didn't fill in NCB_CALLNAME
 *	              properly (did not affect operation).
 *	10/31/90 RAE - Added OS/2 compiler directives.  There are two major
 *	              differences between DOS and OS/2.  In DOS, exec_netbios loads
 *	              registers ES:BX with the address of an NCB, while under OS/2,
 *	              exec_netbios simply calls the DLL library.  In DOS, the NCB
 *	              pool is declared globally in this file (netlib.c), while under
 *	              OS/2, the NCB pool is allocated elsewhere in the driver.  This
 *	              affects both the ncb_reset function, and the ncb_open function.
 *	              Also, OS/2 does not support ncb_sa85off, but does require and
 *	              additional function called ncb_tick().  Also fixed a bug in DOS
 *	              version, when ncb_reset was issued, it cleared the ncbs array
 *	              for BOTH boards.
 *	4/12/90 RAE - Changed MAX_NCB to 64 to support two hosts (max of 64 paths)
 *	3/5/90  RAE - Modified exec_netbios routine to check for null pointer
 *	              interrupt vectors before invoking int86.
 *	1/31/90 RAE - Modified ncb_receive and ncb_receive_wait to add 1 to the
 *	              length returned to account for the modbus device address
 *	              being added to the beginning of the message, also fixed the
 *	              device address by using the last non-zero byte of the routing
 *	              path instead of the first byte of the routing path.
 *	1/31/90 RAE - Removed the length fix above.
 *=fdoc=====================================================================*/

							      

/*
*	include files
*/

#ifdef _WINDLL
#include "windows.h"
#include "dpmi.h"
#endif

#include <stdio.h>
#ifndef __OS2__
#include <dos.h>
#endif
#include <process.h>
#include <string.h>
#include <ctype.h>
#include <stdlib.h>
#include "netbios.h"
#include "netlib.h"

#ifdef __OS2__
    extern void OS2ENTRY mbp_netbios(NCB *abc);

#else /* DOS */

#ifdef _WINDLL	/* Windows DLL */

DWORD FAR PASCAL GlobalDosAlloc(DWORD);	/* Part of Windows Kernel */
UINT FAR PASCAL GlobalDosFree(UINT);	/* Part of Windows Kernel */

#else /* DOS static library */

extern unsigned cdecl far dv_get_version(void);
extern void cdecl far dv_begin_critical(void);
extern void cdecl far dv_end_critical(void);
extern unsigned cdecl far IsEnhancedWindows(void);
extern void cdecl far WinEnterCriticalSection(void);
extern void cdecl far WinEndCriticalSection(void);
#endif /* _WINDLL */
#endif /* OS2 */


/*
**	The following defines deal with the partitioning of the global memory
**	resources. When multiple tasks have access to the DLL it is important that
**	each have equal opportunities to use the resources. Officially, only two
**	adapters are supported. However, the early drivers (3.2 and earlier) and 
**	newer drivers (3.41 and newer), have no restrictions on the number of
**	adapters on the backplane (the /n config.sys option.) IF YOU WISH TO TALK
**	TO MORE THAN TWO ADAPTERS IN THE SYSTEM, YOU MUST CHANGE MAX_ADAPTERS AND
** RECOMPILE THIS LIBRARY/DLL.
*/

#define MAX_NCB 64
#define MAX_ADAPTERS	2
#define NCB_PER_BOARD (MAX_NCB/MAX_ADAPTERS)

/*
**	global variables
*/

#ifndef __OS2__
NCB ncbs[MAX_NCB];              /*use for program,data paths*/
int sw_interrupt = 0x5c;        /*s/w interrupt to use for exec_netbios()*/
NCB dummy_ncb;			/* Dummy ncb for use by netlib */
#endif


/*=doc========================================================================
 *
 * Function:    ncb_reset
 *
 * Description:	Issue reset command to the host adapter board.  This causes all active
 *						transactions to be cleared, and all paths to be closed.
 *						USE WITH CAUTION!
 *
 * Returns:		The return code of the NCB structure.
 *
 * Parameters:  Adapter number.
 *
 * Notes:       
 *
 * Revision History Log:
 *
 *	NAME		DATE			COMMENT
 *	-----		-----			--------
 * DP			04/14/94		Merged DOS and OS/2 with Windows DLL code.
 *=doc======================================================================*/

int far ncb_reset (int adaptno)
{
#ifndef __OS2__
    int i;				
    NCB  far *ncbp;
#endif
    extern NCB far dummy_ncb;

    if ((adaptno < 0) || (adaptno >= MAX_ADAPTERS))
        return ERR_invalid;
    dummy_ncb.NCB_LANA_NUM = (unsigned char) adaptno;
    dummy_ncb.NCB_COMMAND = RESET;
                                        /*invalidate return code*/
    dummy_ncb.NCB_RETCODE = ERR_bad_card_num;
    dummy_ncb.NCB_LSN = 0;
    dummy_ncb.NCB_POST_ADDRESS = NULL;
    exec_netbios (&dummy_ncb);
#ifndef __OS2__
    for (i = 0;  i < NCB_PER_BOARD;  i++) 
    {
        /* reset them all to unused */
        ncbp = (NCB *) &ncbs[i + (NCB_PER_BOARD * adaptno)];
        ncbp->NCB_NUM = 0;
    }
#endif
    return dummy_ncb.NCB_RETCODE;
}

/*=doc========================================================================
 *
 * Function:    ncb_sa85off
 *
 * Description: 	Turn off SA85 driver (RESET to turn back on).  DOS only.
 *
 * Returns:    	NCB return code.
 *
 * Parameters:		Adapter Number
 *
 * Notes:       
 *
 * Revision History Log:
 *
 *	NAME		DATE			COMMENT
 *	-----		-----			--------
 *	DP			04/14/94		Merged DOS and OS/2 code with Windows DLL.
 *=doc======================================================================*/

#ifndef __OS2__

int far ncb_sa85off (int adaptno)
{

    if ((adaptno < 0) || (adaptno >= MAX_ADAPTERS))
        return ERR_invalid;

    dummy_ncb.NCB_LANA_NUM = (unsigned char) adaptno;
    dummy_ncb.NCB_COMMAND = SA85_OFF;
                                        /*invalidate return code*/
    dummy_ncb.NCB_RETCODE = ERR_bad_card_num;
    exec_netbios(&dummy_ncb);
    return dummy_ncb.NCB_RETCODE;
}
#endif


/*=doc========================================================================
 *
 * Function:   	ncb_status()
 *
 * Description: 	Configuration status. Performs a quick test of the shared
 *						ram interface.
 *
 * Returns:     	NCB return code.
 *
 * Parameters:  Pointer to NCB structure and an adapter number.
 *
 * Notes:       x
 *
 * Revision History Log:
 *
 *	NAME		DATE			COMMENT
 *	-----		-----			--------
 *	DP			04/14/94		Merged DOS and OS/2 code with Windows DLL.
 *=doc======================================================================*/


int far ncb_status (NCB far * ncbp, int adaptno)
{
    if ((adaptno < 0) || (adaptno >= MAX_ADAPTERS))
        return (ncbp->NCB_RETCODE = ERR_invalid);

    ncbp->NCB_LANA_NUM = (unsigned char) adaptno;
    ncbp->NCB_COMMAND = STATUS;
                                        /*invalidate return code*/
    ncbp->NCB_RETCODE = ERR_bad_card_num;
    ncbp->NCB_LSN = 0;
    ncbp->NCB_POST_ADDRESS = NULL;
    exec_netbios (ncbp);
    return ncbp->NCB_RETCODE;
}
     

/*=doc========================================================================
 *
 * Function:    ncb_send()
 *
 * Description: 	Send a message on the given path.
 *
 * Returns:     Returns NCB return code value.												 
 *
 * Parameters:  	Pointer to NCB structure.
 *						Length of user buffer.
 *						Pointer to user buffer.
 *						Timeout value.
 *
 * Notes:       
 *
 * Revision History Log:
 *
 *	NAME		DATE			COMMENT
 *	-----		-----			--------
 *	DP			04/14/94		Merged DOS and OS/2 code with Windows DLL.
 *=doc======================================================================*/

int far ncb_send (NCB far *ncbp, int length, char far *buffer,
																unsigned char timeout)
{
   if (ncbp->NCB_NUM == 0)             /*verify path exists*/
	{
		return (ncbp->NCB_RETCODE = ERR_bad_name_num);
	}
   ncbp->NCB_LENGTH = length - 1;      /*copy length to ncb*/
   ncbp->NCB_BUFFER = buffer + 1;      /*copy buffer address to ncb*/
   ncbp->NCB_STO = timeout;            /*copy timeout to ncb*/
   ncbp->NCB_COMMAND = SEND;           /*send command*/
                                        /*invalidate return code*/
   ncbp->NCB_RETCODE = ERR_bad_card_num;
   ncbp->NCB_LSN = 0;
   ncbp->NCB_POST_ADDRESS = NULL;
   exec_netbios (ncbp);                /*execute*/
   return ncbp->NCB_RETCODE;
}

/*=doc========================================================================
 *
 * Function:    ncb_cancel()
 *
 * Description: 	Cancel an outstanding transaction.
 *
 * Returns:     Returns NCB return code value.
 *
 * Parameters:  Pointer to NCB structure.
 *
 * Notes:       x
 *
 * Revision History Log:
 *
 *	NAME		DATE			COMMENT
 *	-----		-----			--------
 *	DP			04/14/94		Merged DOS and OS/2 code with Windows DLL.
 *=doc======================================================================*/


int far ncb_cancel (NCB far *ncbp)
{
    printf( "CANCEL CALLED\n" );
    if (ncbp->NCB_NUM == 0)             /*verify path exists*/
        return -1;
    ncbp->NCB_COMMAND = CANCEL;         /*send command*/
                                        /*invalidate return code*/
    ncbp->NCB_RETCODE = ERR_bad_card_num;
    ncbp->NCB_LSN = 0;
    ncbp->NCB_POST_ADDRESS = NULL;
    exec_netbios (ncbp);                /*execute*/
    return ncbp->NCB_RETCODE;
}

/*=doc========================================================================
 *
 * Function:    ncb_receive()
 *
 * Description: 	Receive a message on the given path.
 *
 * Returns:     Returns NCB return code value.
 *
 * Parameters:  	Pointer to NCB structure.
 *						Pointer the user buffer.
 *
 * Notes: 		ncb_receive() has changed for rev 2.  For rev 1, it would wait
 *					unitl the response came in, or it timed out.
 *					Now, ncb_receive_wait() does that. If the response is not yet 
 *					available when ncb_receive() is called, it returns immediately
 *					with the ERR_timeout status. This makes it more useful in a 
 *					multi-tasking environment.
 *
 *					When the operation is complete, the length of the received 
 *					packet is contained in NCB_LENGTH.
 *
 * Revision History Log:
 *
 *	NAME		DATE			COMMENT
 *	-----		-----			--------
 *	DP			04/14/94		Merged DOS and OS/2 code with Windows DLL.
 *=doc======================================================================*/

int far ncb_receive (NCB far *ncbp, char far *buffer)
{
    int    i;

    if (ncbp->NCB_NUM == 0)             /*verify path exists*/
        return -1;
    ncbp->NCB_BUFFER = buffer + 1;      /*copy buffer address to ncb*/
    ncbp->NCB_RTO = 0;                  /* not used for RECEIVE */
    ncbp->NCB_COMMAND = RECEIVE;        /* rev 1 driver used RECEIVE command*/
                                        /*invalidate return code*/
    ncbp->NCB_RETCODE = ERR_bad_card_num;
    ncbp->NCB_LSN = 0;
    exec_netbios (ncbp);                /*execute*/
    i = 0;
    while( ncbp->NCB_CALLNAME[ 2 + i ] && ( i < 5 ) )
        i++;
    if( i )
        i--;                            /* back up to last non-zero byte */
    buffer[ 0 ] = ncbp->NCB_CALLNAME[ 2 + i ];

    return ncbp->NCB_RETCODE;									
}

/*=doc========================================================================
 *
 * Function:    ncb_receive_wait()
 *
 * Description: 	Receive a message on the given path.
 *
 * Returns:     Returns NCB return code value.
 *
 * Parameters:  	Pointer to NCB structure.
 *						Pointer the user buffer.
 *						timeout value.
 *
 * Notes:      	This is the same as the rev 1 receive() function.
 *						See the comments for ncb_receive() as well.
 *
 * Revision History Log:
 *
 *	NAME		DATE			COMMENT
 *	-----		-----			--------
 *	DP			04/14/94		Merged DOS and OS/2 code with Windows DLL.
 *=doc======================================================================*/


int far ncb_receive_wait(NCB far *ncbp, char far *buffer, unsigned char timeout)
{
    int    i;

    if (ncbp->NCB_NUM == 0)             /*verify path exists*/
	    return (ncbp->NCB_RETCODE = ERR_bad_card_num);

    ncbp->NCB_BUFFER = buffer + 1;      /*copy buffer address to ncb*/
    ncbp->NCB_RTO = timeout;            /*copy timeout to ncb*/
    ncbp->NCB_COMMAND = RECEIVE_WAIT;   /*receive command, rev 2 driver*/
                                        /*invalidate return code*/
    ncbp->NCB_RETCODE = ERR_bad_card_num;
    ncbp->NCB_LSN = 0;
    ncbp->NCB_POST_ADDRESS = NULL;
    exec_netbios (ncbp);                /*execute*/
    i = 0;
    while( ncbp->NCB_CALLNAME[ 2 + i ] && ( i < 5 ) )
        i++;
    if( i )
        i--;                            /*backup to last non-zero byte*/
    buffer[ 0 ] = ncbp->NCB_CALLNAME[ 2 + i ];
    return ncbp->NCB_RETCODE;
}

/*=doc========================================================================
 *
 * Function:    ncb_open()
 *
 * Description: Open the specified path
 *
 * Returns:     Return pointer to NCB, or NULL on error.
 *
 * Parameters:  	Path name.
 *						Pointer to NCB structure.
 *
 * Notes:       x
 *
 * Revision History Log:
 *
 *	NAME		DATE			COMMENT
 *	-----		-----			--------
 *	DP			04/14/94		Merged DOS and OS/2 code with Windows DLL.
 *=doc======================================================================*/


#ifdef __OS2__

NCB * ncb_open (char * name, int adaptno)
{
    int i, j, k;
    NCB temp_ncb, *temp_ncbp = &temp_ncb;
    unsigned char temp_path[2];
    unsigned char *ucp;
    unsigned char r[5];
    char *cptr = name;
    unsigned char *ucptr = r;

    if ((adaptno < 0) || (adaptno >= MAX_ADAPTERS))
         return (NCB *) NULL;

    /*
     * First, check to see if it's for a Slave path:
     * Example: ncb_open("DS.1.0.0.0.0",0);
     *        open slave data path #1 (1-8) (internal 0x41-0x48), 
     *        lan adapter #0
     *        (for Slave paths, the path number must be explicitly
     *         chosen, but no other routing information is specified )
     */

    if (toupper(name[1]) == 'S') 
    {
        temp_path[0] = toupper(name[3]);
        temp_path[1] = 0;
        if (toupper(name[0]) == 'D')
            temp_ncbp->NCB_NUM = (unsigned char) (0x40 + (atoi(temp_path)));
        else 
            temp_ncbp->NCB_NUM = (unsigned char) (0xC0 + (atoi(temp_path)));

        temp_ncbp->NCB_CALLNAME[0] = name[0];                 /*'D' or 'P'*/
        temp_ncbp->NCB_CALLNAME[1] = name[1];                 /*'S'    'S'*/
        temp_ncbp->NCB_CALLNAME[2] = temp_ncbp->NCB_NUM & 0xf;/* 1      1 */
        temp_ncbp->NCB_CALLNAME[3] = 0;                       /* 0      0 */
        temp_ncbp->NCB_CALLNAME[4] = 0;                       /* 0      0 */
        temp_ncbp->NCB_CALLNAME[5] = 0;                       /* 0      0 */
        temp_ncbp->NCB_CALLNAME[6] = 0;                       /* 0      0 */

        temp_ncbp->NCB_COMMAND = ADD_NAME;
        temp_ncbp->NCB_LANA_NUM = (unsigned char) adaptno;
        temp_ncbp->NCB_LSN = 0;
        temp_ncbp->NCB_POST_ADDRESS = NULL;
        exec_netbios (temp_ncbp);

        /*
         * Explanation:
         * 1) User fills in dummy NCB
         * 2) user calls add_name, gets assigned a real NCB
         * 3) driver copies user's values into real NCB
         * 4) drivers stuffs address of real NCB into NCB_BUFFER entry
         * 5) back in netlib.c's ncb_open routine, real pointer to NCB is obtained
         */

        if (temp_ncbp->NCB_RETCODE == 0)
                                        /*return pointer to real NCB */
            return (NCB *) temp_ncbp->NCB_BUFFER;
        else
            return (NCB *) NULL;        /*return null pointer*/
    }

    /*
     * else, it must be for a Master path:
     * Example: ncb_open("DM.13.5.0.0.0",0);
     *        open master data path to node 5 going through router
     *        at node 13,
     *        lan adapter #0
     *        (internal path used for master paths is automatically
     *         selected from those not in use)
     */

    else 
    {
        k = 5;
        j = strlen(name);
        if (j < 2)                      /*types limited to 2*/
            return ((NCB *) NULL);
        i = 0;
        while (!(isdigit(*cptr)) && *cptr) 
        {
            cptr++;
            i++;
        }
        *ucptr = '\0';
        while (*cptr && i < j && k != 0) 
        {
            if (isdigit(*cptr)) 
            {
                *ucptr = ((*ucptr) * 10) + *cptr - '0';
            }
            else 
            {
                ucptr++;
                *ucptr = '\0';
                k--;
            }
            cptr++;
            i++;
        }
        ucp = temp_ncbp->NCB_CALLNAME;
        *ucp++ = *name;
        *ucp++ = *(name + 1);           /*fill in path*/
        *ucp++ = r[0];                  /*convert to binary*/
        *ucp++ = r[1];
        *ucp++ = r[2];
        *ucp++ = r[3];
        *ucp++ = r[4];                  /*fill in routing info*/
        temp_ncbp->NCB_COMMAND = ADD_NAME;
        temp_ncbp->NCB_LANA_NUM = (unsigned char) adaptno;
        temp_ncbp->NCB_LSN = 0;
        temp_ncbp->NCB_POST_ADDRESS = NULL;
        exec_netbios(temp_ncbp);
        if (temp_ncbp->NCB_RETCODE == 0)
                                        /*return pointer to real NCB */
            return (NCB *) temp_ncbp->NCB_BUFFER;
        else 
            return (NCB *) NULL;        /*return null pointer*/
    }
}

#else /**************** Regular DOS ncb_open start here ***************/

NCB far * far ncb_open(char far *name, int adaptno)
{
    int i, j, k;
    NCB far *ncbp;
    unsigned char far *ucp;
    char far *cptr = name;
    unsigned char r[5];
    unsigned char temp_path[2];
    unsigned char far *ucptr = r;

    if ((adaptno < 0) || (adaptno >= MAX_ADAPTERS))
        return (NCB *) NULL;

    /*
     * First, check to see if it's for a Slave path:
     * Example: ncb_open("DS.1.0.0.0.0",0);
     *        open slave data path #1 (1-8) (internal 0x41-0x48), 
     *        lan adapter #0
     *        (for Slave paths, the path number must be explicitly
     *         chosen, but no other routing information is specified )
     */

    if (toupper(name[1]) == 'S') 
    {
        for (i = 0;  i < MAX_NCB;  i++) /*look for first open NCB*/
        {
            ncbp = (NCB *) &ncbs[i];
            if (ncbp->NCB_NUM == 0)     /*use first available*/
            {
                temp_path[0] = toupper(name[3]);
                temp_path[1] = 0;
                if (toupper(name[0]) == 'D')
                    ncbp->NCB_NUM = (unsigned char) (0x40 + (atoi(temp_path)));
                else 
                    ncbp->NCB_NUM = (unsigned char) (0xC0 + (atoi(temp_path)));

                ncbp->NCB_CALLNAME[0] = name[0];                /*'D' or 'P'*/
                ncbp->NCB_CALLNAME[1] = name[1];                /*'S'    'S'*/
                ncbp->NCB_CALLNAME[2] = ncbp->NCB_NUM & 0xf;    /* 1      1 */
                ncbp->NCB_CALLNAME[3] = 0;                      /* 0      0 */
                ncbp->NCB_CALLNAME[4] = 0;                      /* 0      0 */
                ncbp->NCB_CALLNAME[5] = 0;                      /* 0      0 */
                ncbp->NCB_CALLNAME[6] = 0;                      /* 0      0 */
                ncbp->NCB_COMMAND = ADD_NAME;
                ncbp->NCB_LANA_NUM = (unsigned char) adaptno;
                                        /*invalidate return code*/
                ncbp->NCB_RETCODE = ERR_bad_card_num;
                exec_netbios (ncbp);
                if (ncbp->NCB_RETCODE == 0)
                    return ncbp;        /*return pointer to NCB*/
                else
					 {	  
						  ncbp->NCB_NUM = 0;	 /* Flag as available */
                    return (NCB *) NULL;/*return null pointer*/
					 }
            }
        }
        return (NCB *) NULL;            /*return null pointer*/
    }

    /*
     * else, it must be for a Master path:
     * Example: ncb_open("DM.13.5.0.0.0",0);
     *        open master data path to node 5 going through router
     *        at node 13,
     *        lan adapter #0
     *        (internal path used for master paths is automatically
     *         selected from those not in use)
     */

    else 
    {
        k = 5;
        j = strlen(name);
        if (j < 2)                      /*types limited to 2*/
            return ((NCB *) NULL);
        i = 0;
        while (!(isdigit(*cptr)) && *cptr) 
        {
            cptr++;
            i++;
        }
        *ucptr = '\0';
        while (*cptr && i < j && k != 0) 
        {
            if (isdigit(*cptr)) 
            {
                *ucptr = ((*ucptr) * 10) + *cptr - '0';
            }
            else 
            {
                ucptr++;
                *ucptr = '\0';
                k--;
            }
            cptr++;
            i++;
        }

        for (i = 0;  i < MAX_NCB;  i++) /*look for first open NCB*/
        {
            ncbp = (NCB *) &ncbs[i];
            if (ncbp->NCB_NUM == 0)     /*use first available*/
            {
                ucp = ncbp->NCB_CALLNAME;
                *ucp++ = *name;
                *ucp++ = *(name + 1);   /*fill in path*/
                *ucp++ = r[0];          /*convert to binary*/
                *ucp++ = r[1];
                *ucp++ = r[2];
                *ucp++ = r[3];
                *ucp++ = r[4];          /*fill in routing info*/
                ncbp->NCB_COMMAND = ADD_NAME;
                ncbp->NCB_LANA_NUM = (unsigned char) adaptno;
                                        /*invalidate return code*/
                ncbp->NCB_RETCODE = ERR_bad_card_num;
                exec_netbios (ncbp);
                if (ncbp->NCB_RETCODE == 0)
                    return ncbp;        /*return pointer to NCB*/
                else
					 {	  
						  ncbp->NCB_NUM = 0;	 /* Flag as available */
                    return (NCB *) NULL;/*return null pointer*/
					 }
            }
        }
        return (NCB *) NULL;            /*return null pointer*/
    }
}

#endif

/*=doc========================================================================
 *
 * Function:    ncb_close()
 *
 * Description: Close a path. 
 *
 * Returns:     Returns NCB return code value.
 *
 * Parameters:  Pointer to NCB structure.
 *
 * Notes:       x
 *
 * Revision History Log:
 *
 *	NAME		DATE			COMMENT
 *	-----		-----			--------
 *	DP			04/14/94		Merged DOS and OS/2 code with Windows DLL.
 *=doc======================================================================*/

int  far ncb_close(NCB far * ncbp)
{
#ifdef __OS2__
    NCB temp_ncb;
#endif
    ncbp->NCB_COMMAND = DELETE_NAME;        
                                        /*invalidate return code*/
    ncbp->NCB_RETCODE = ERR_bad_card_num;
    ncbp->NCB_LSN = 0;                      
    ncbp->NCB_POST_ADDRESS = NULL;          

#ifdef __OS2__
    temp_ncb = *ncbp;
    temp_ncb.NCB_BUFFER = (char far *) ncbp;

    exec_netbios (&temp_ncb);

    temp_ncb.NCB_NUM = 0;
    return (int) temp_ncb.NCB_RETCODE;
#else
    exec_netbios(ncbp);
    ncbp->NCB_NUM = 0;  /* only if good status */
    return (int) ncbp->NCB_RETCODE;
#endif
}

/*=doc========================================================================
 *
 * Function:    ncb_send_datagram()
 *
 * Description: 	Write global data from this station.  This data will be 
 *						broadcast to all other stations on the local link every time
 *						this station passes the token.
 *
 * Returns:     Returns NCB return code value.
 *
 * Parameters:  	Pointer to NCB structure.
 *						Length of global data.
 *						Pointer to the user buffer.
 *						Timeout value.
 *						Adapter number.
 *
 * Notes:       
 *
 * Revision History Log:
 *
 *	NAME		DATE			COMMENT
 *	-----		-----			--------
 *	DP			04/14/94		Merged DOS and OS/2 code with Windows DLL.
 *=doc======================================================================*/

int  far ncb_send_datagram(NCB far * ncbp, int length, char * far buffer, 
                        unsigned char timeout, int adaptno)
{
    if (length > 64)
        return ERR_bad_buffer_length;
    ncbp->NCB_LANA_NUM = (unsigned char) adaptno;
    ncbp->NCB_COMMAND = SEND_DATAGRAM;
    ncbp->NCB_BUFFER = buffer;
    ncbp->NCB_STO = timeout;            /*copy timeout to ncb*/
    if (length & 1)
                                        /*convert to words*/
        ncbp->NCB_LENGTH = (length / 2) + 1;
    else 
        ncbp->NCB_LENGTH = length / 2;  /*convert to words*/
                                        /*invalidate return code*/
    ncbp->NCB_RETCODE = ERR_bad_card_num;
    ncbp->NCB_LSN = 0;
    ncbp->NCB_POST_ADDRESS = NULL;
    exec_netbios (ncbp);
    return (int) ncbp->NCB_RETCODE;
}

/*=doc========================================================================
 *
 * Function:    ncb_receive_datagram()
 *
 * Description: 	Read global data from another station on the local link.  
 *						The data read was written to the local stations memory the 
 *						last time the station being read from passed the token.
 *						The length in words is returned in NCB_LENGTH.
 *
 * Returns:     Returns NCB return code value.
 *
 * Parameters:  	Pointer to NCB structure.
 *						Remote node's address.
 *						Pointer to the user buffer.
 *						Timeout value.
 *						Adapter number.
 *
 * Notes:       x
 *
 * Revision History Log:
 *
 *	NAME		DATE			COMMENT
 *	-----		-----			--------
 *	DP			04/14/94		Merged DOS and OS/2 code with Windows DLL.
 *=doc======================================================================*/


int  far ncb_receive_datagram(NCB far *ncbp, int node, char far *buffer, 
                        unsigned char timeout, int adaptno)
{
    ncbp->NCB_LANA_NUM = (unsigned char) adaptno;
    ncbp->NCB_COMMAND = RECEIVE_DATAGRAM;   /*get global data*/
    ncbp->NCB_NUM = (unsigned char) node;
    ncbp->NCB_BUFFER = buffer;
    ncbp->NCB_RTO = timeout;                /*copy timeout to ncb*/
                                        /*invalidate return code*/
    ncbp->NCB_RETCODE = ERR_bad_card_num;
    ncbp->NCB_LSN = 0;
    ncbp->NCB_POST_ADDRESS = NULL;
    exec_netbios (ncbp);
    return ncbp->NCB_RETCODE;
}


/*=doc========================================================================
 *
 * Function:    ncb_set_slave_login()
 *
 * Description: 	Set login status on a slave path.
 *						login_status = 1 = set login status
 *						login_status = 0 = clear login status
 *
 * Returns:     Returns NCB return code value.
 *
 * Parameters:  Pointer to NCB structure.
 *
 * Notes:    	The path must have been previously opened. Closing the path 
 *					resets the login status to not logged in.
 *
 * Revision History Log:
 *
 *	NAME		DATE			COMMENT
 *	-----		-----			--------
 *	DP			04/14/94		Merged DOS and OS/2 code with Windows DLL.
 *=doc======================================================================*/

int far ncb_set_slave_login(NCB far *ncbp, unsigned char login_status)
{
    unsigned char buffer[2];

    if (ncbp->NCB_NUM == 0)             /*verify path exists*/
        return (ncbp->NCB_NUM = ERR_bad_name_num);

    ncbp->NCB_BUFFER = buffer;          /*copy buffer address to ncb*/
    ncbp->NCB_BUFFER[0] = login_status;
    ncbp->NCB_COMMAND = SET_SLAVE_LOGIN;/*send command*/
                                        /*invalidate return code*/
    ncbp->NCB_RETCODE = ERR_bad_card_num;
    ncbp->NCB_LSN = 0;
    ncbp->NCB_POST_ADDRESS = NULL;
    exec_netbios (ncbp);                /*execute*/
    return ncbp->NCB_RETCODE;
}

/*=doc========================================================================
 *
 * Function:    ncb_set_sw_interrupt()
 *
 * Description: Set the software interrupt number.
 *
 * Returns:  	    ERR_invalid - illegal value
 *					    ERR_success - successfully changed
 *
 * Parameters:  Pointer to NCB structure.
 *
 * Notes:       
 *
 * Revision History Log:
 *
 *	NAME		DATE			COMMENT
 *	-----		-----			--------
 *	DP			04/14/94		Merged DOS and OS/2 code with Windows DLL.
 *=doc======================================================================*/


int far ncb_set_sw_interrupt(int swInterrupt)
{
    if ((swInterrupt < 3) ||
        ((swInterrupt >= 8) && (swInterrupt <= 0xf)) ||
        (swInterrupt > 0xff))
    {
        return ERR_invalid;
    }
    sw_interrupt = swInterrupt;
    return ERR_success;
}

/*=doc========================================================================
 *
 * Function:    exec_netbios()
 *
 * Description: Execute a NetBIOS command via the software interrupt.
 *
 * Returns:     None
 *
 * Parameters:  Pointer to NCB structure.
 *
 * Notes:     	OS/2 version. Here the software interrupt is replaced with a
 *					call to the dynamic link library.  
 *
 * Revision History Log:
 *
 *	NAME		DATE			COMMENT
 *	-----		-----			--------
 *=doc======================================================================*/

#ifdef __OS2__

void exec_netbios(NCB far * ncbstruct)
{
    mbp_netbios(ncbstruct);             /*go do it in DLL*/
    return;
}

#else   /* DOS and Windows. */

/*=doc========================================================================
 *
 * Function:    exec_netbios()
 *
 * Description: Execute a NetBIOS command via the software interrupt.
 *
 * Returns:     None
 *
 * Parameters:  Pointer to NCB structure.
 *
 * Notes:       This is the Windows DLL version.
 *
 * Revision History Log:
 *
 *	NAME		DATE			COMMENT
 *	-----		-----			--------
 * DP			01/09/95		Return ERR_sess_table_full if there is no memory.
 *=doc======================================================================*/

#ifdef _WINDLL  /* Windows DLL */


DPMIXLATSTRUCT DpmiRegs; /* Translate struct for dpmi interface */

void exec_netbios(NCB far * ncbstruct)
{

	NCB FAR *lpProtectedNCB; /* Protected mode selector */
	LPSTR lpUserBuffer;		/* Alias for users buffer */

	DWORD dwNetBiosBuffer;	/* Handle of dos accessable memory */
	NCB FAR *lpDosNCB;		/* Dos accessable segment:offset */

	extern DWORD dwWinFlags;
	extern int fIsEnhanced;

	if ((dwWinFlags & WF_PMODE))
   {
   	if ((dwNetBiosBuffer = GlobalDosAlloc((DWORD) sizeof (NCB) + 1021L))
					== (DWORD) NULL)
	   {
			ncbstruct->NCB_RETCODE = ERR_sess_table_full;
			return;
      }
		FP_OFF(lpProtectedNCB) = 0;
		FP_SEG(lpProtectedNCB) = LOWORD(dwNetBiosBuffer);	/* use selector value */
		FP_OFF(lpDosNCB) = 0;
		FP_SEG(lpDosNCB) = HIWORD(dwNetBiosBuffer);			/* use segment value */
		lpProtectedNCB->NCB_BUFFER = (LPSTR) (lpDosNCB + 1);
		DpmiRegs.ES = HIWORD(dwNetBiosBuffer);
		DpmiRegs.EBX = 0L;

    /* ====================================================================
			    PROTECTED MODE
	      TRANSFER CALLERS NCB TO DOS ACCESSABLE NCB
       ==================================================================== */

		lpUserBuffer = lpProtectedNCB->NCB_BUFFER; /* Save data buffer */
		_fmemcpy((LPSTR) lpProtectedNCB, (LPSTR) ncbstruct, sizeof (NCB));
		lpProtectedNCB->NCB_BUFFER = lpUserBuffer; /* Restore data buffer */


    /* ====================================================================
		      IF COMMAND HAS DATA PACKET
		TRANSFER DATA TO DOS ACCESSABLE BUFFER
       ==================================================================== */

		switch (ncbstruct->NCB_COMMAND)
      {
      	case SEND:
		      _fmemcpy((LPSTR) (lpProtectedNCB + 1), (LPSTR) ncbstruct->NCB_BUFFER,
	   		 ncbstruct->NCB_LENGTH);
	      	break;
	
   	   case SEND_DATAGRAM:
      		_fmemcpy((LPSTR) (lpProtectedNCB + 1), (LPSTR) ncbstruct->NCB_BUFFER,
		      ncbstruct->NCB_LENGTH * 2);
   		   break;
	
	      case SET_SLAVE_LOGIN:
   		   _fmemcpy((LPSTR) (lpProtectedNCB + 1), (LPSTR) ncbstruct->NCB_BUFFER, 2);
		      break;

      	default:
	      break;
      }

    /* ====================================================================
			      CALL DPMI
       ==================================================================== */
    
		asm  cmp	fIsEnhanced,0		// If 386 Enhanced mode
		asm  jz	No386
		asm  mov	ax,1681h		// Begin Critical Section
		asm  int	2fh
    No386:
		asm  push	di			// Save used register
	   asm  mov	bx,sw_interrupt		// Interrupt number
   	asm  xor	bh,bh			// Flags
	   asm  xor	cx,cx			// Parameter count (0 in this case)
   	asm  push	ds			// ES:DI pointer to translate
	   asm  pop	es			// ...structure
   	asm  mov	di,offset DpmiRegs	// Pointer to translate structure
	   asm  mov	ax,0300h		// Simulate Interrupt request
   	asm  int	31h			// Call DPMI
	   asm  pop	di			// Restore register
   	asm  cmp	fIsEnhanced,0		// If 386 Enhanced mode
	   asm  jz	ExitDPMI	
   	asm  mov	ax,1682h		// End Critical Section
	   asm  int	2fh
   ExitDPMI:
    /* ====================================================================
		    TRANSFER NCB TO CALLERS BUFFER
       ==================================================================== */

		lpUserBuffer = ncbstruct->NCB_BUFFER; /* Save callers pointer */
	   _fmemcpy((LPSTR) ncbstruct, (LPSTR) lpProtectedNCB, sizeof(NCB));
   	ncbstruct->NCB_BUFFER = lpUserBuffer; /* Restore callers pointer */

    /* ====================================================================
	     IF NCB HAS DATA COPY DATA TO CALLERS BUFFER
       ==================================================================== */
	   if( ncbstruct->NCB_RETCODE == ERR_success )
	   {
   	   switch (ncbstruct->NCB_COMMAND)
      	{
            case RECEIVE:
            case RECEIVE_WAIT:
                _fmemcpy((LPSTR) ncbstruct->NCB_BUFFER, (LPSTR) (lpProtectedNCB + 1),
	                ncbstruct->NCB_LENGTH);
                break;

	        case STATUS:
	            _fmemcpy((LPSTR) ncbstruct->NCB_BUFFER, (LPSTR) (lpProtectedNCB + 1),
			        110);
	            break;

            case RECEIVE_DATAGRAM:
                _fmemcpy((LPSTR) ncbstruct->NCB_BUFFER, (LPSTR) (lpProtectedNCB + 1),
	                ncbstruct->NCB_LENGTH * 2);
                break;

               default:
                break;
      	}
    	}
	   GlobalDosFree(FP_SEG(lpProtectedNCB));
   	lpProtectedNCB = NULL;
	}
	else
    /* ====================================================================
			      REAL MODE
       ==================================================================== */
   {
	   DWORD dwIntVector;
    
   	asm  xor	bx,bx			//Get interrupt vector from
	   asm  mov	es,bx			//...0:sw_interrupt *4
   	asm  mov	bx,sw_interrupt
	   asm  shl	bx,1
   	asm  shl	bx,1
		asm  mov	ax,es:[bx]		// Save Interrupt vector to variable
	   asm  mov	word ptr dwIntVector+0,ax
   	asm  mov	ax,es:[bx + 2]
	   asm  mov	word ptr dwIntVector+2,ax
   	asm  push	ds
	   asm  pop	es
   	asm  mov	bx,offset ncbstruct		// ES:BX points to NCB
	   asm  pushf				// Simulate an interrupt
   	asm  call	dword ptr dwIntVector
    
	}
	   
}

/*=doc========================================================================
 *
 * Function:    exec_netbios()
 *
 * Description: Execute a NetBIOS command via the software interrupt.
 *
 * Returns:     None
 *
 * Parameters:  Pointer to NCB structure.
 *
 * Notes:   	DOS version. Critical regions for both Windows 3.x and 
 *					DESQview are placed around the entrypoint to the driver.  
 *
 * Revision History Log:
 *
 *	NAME		DATE			COMMENT
 *	-----		-----			--------
 *=doc======================================================================*/

#else	/* DOS */	

void exec_netbios(NCB far * ncbstruct)
{
	 int fIsEnhancedWindows;
    union REGS inregs, outregs;
    struct SREGS segregs;
    int result_ax;
    unsigned long far *ivector;

    /*
     * Note, dv_get_version squirrels away in its own variable whether
     * DESQviews there or not, so we don't need to qualify the
     * call to dv_begin_critical() or dv_end_critical()
     */

    dv_get_version();                   /*see if DESQview is there*/
    dv_begin_critical();                /*begin DESQview critical area*/

    inregs.x.bx = FP_OFF (ncbstruct);   /*BX = offset of structure*/
    segregs.es = FP_SEG (ncbstruct);    /*ES = segment of pointer*/

    /*
     * Note: The default value for sw_interrupt is 0x5c, the traditional
     * value for the NetBIOS s/w interrupt.
     * If you change it, the values in the global variable "sw_interrupt"
     * and the setting for the sa85.sys driver in config.sys must match.
     */

    ivector = (unsigned long far *) (long) (sw_interrupt * 4);

    if (fIsEnhancedWindows = IsEnhancedWindows())
        WinEnterCriticalSection();      /*enter windows critical section*/

    if(*ivector)                        /*look for non-null interrupt vector*/
        int86x(sw_interrupt,/*software interrupt number*/
            &inregs,                    /*array of registers input*/
            &outregs,                   /*array of registers output*/
            &segregs);                  /*image of segment registers*/

    if (fIsEnhancedWindows)
        WinEndCriticalSection();        /*exit windows critical section*/

    dv_end_critical();                  /*end DESQview critical area*/

}
#endif /* WINDLL */
#endif /* OS2 */


